-- Let It Snow! Shadow of Mal'Ganis US
LetItSnow = CreateFrame("Frame")
LibStub("AceTimer-3.0"):Embed(LetItSnow)

local SNOWFLAKES = {"Interface\\AddOns\\LetItSnow\\images\\snowflake1", "Interface\\AddOns\\LetItSnow\\images\\snowflake2", "Interface\\AddOns\\LetItSnow\\images\\snowflake3", "Interface\\AddOns\\LetItSnow\\images\\snowflake4", "Interface\\AddOns\\LetItSnow\\images\\snowflake5", "Interface\\AddOns\\LetItSnow\\images\\snowflake6"}
local SCREEN_WIDTH = GetScreenWidth()
local SCREEN_HEIGHT = GetScreenHeight()

local hasSnowflakesBuff, scheduledFall, isInCombat
local inactiveFlakes, activeFlakes = {}, {}

local function OnFinished(self)
	if( self.parent.suppress ) then return end
		
	for i=#(activeFlakes), 1, -1 do
		if( activeFlakes[i] == self.parent ) then
			table.insert(inactiveFlakes, table.remove(activeFlakes, i))

			self.parent.animGroup:Stop()
			self.parent.rotateGroup:Stop()
			if( self.quickFadeGroup ) then self.quickFadeGroup:Stop() end
			
			self.parent:Hide()
			break
		end
	end
end

local function OnFade(self)
	self.parent.rotateGroup:Pause()
end

local function getFlake()
	local frame = table.remove(inactiveFlakes, 1)
	if( not frame ) then
		frame = CreateFrame("Frame", nil, UIParent)
		frame:SetFrameStrata("BACKGROUND")
		frame:SetFrameLevel(1)
		
		frame.flake = frame:CreateTexture(nil, "BACKGROUND")
		frame.flake:SetAllPoints(frame)
		
		frame.animGroup = frame:CreateAnimationGroup()
		frame.animGroup.parent = frame
		frame.animGroup:SetLooping("NONE")
		
		frame.rotateGroup = frame:CreateAnimationGroup()
		frame.rotateGroup:SetLooping("REPEAT")
		frame.rotateAnim = frame.rotateGroup:CreateAnimation("Rotation")		
	
		frame.fallAnim = frame.animGroup:CreateAnimation("Translation")
		frame.fadeAnim = frame.animGroup:CreateAnimation("Alpha")
		frame.fadeAnim:SetScript("OnPlay", OnFade)
		frame.fadeAnim.parent = frame
		


		frame.fadeAnim.parent = frame
		frame.fadeAnim:SetScript("OnFinished", OnFinished)
	end
	
	table.insert(activeFlakes, frame)
	return frame
end

local frameList, validFrames, verifiedPosition = {}, {}, {}
local function scanObjects(format)
	local id = 1
	while( true ) do
		local button = _G[string.format(format, id)]
		if( not button ) then return end
		frameList[button] = true
		id = id + 1
	end
end

-- Not exactly clean, but it works quite well in adding any frame using ShowUIPanel including custom mods
hooksecurefunc("ShowUIPanel", function(frame) frameList[frame] = true end)

function LetItSnow:CreateList()
	if( self.initialLoad ) then return end
	self.initialLoad = true
	
	table.wipe(frameList)
	for frame, enabled in pairs(self.db.collisionObjects) do
		if( enabled and string.match(frame, "%%d") ) then
			scanObjects(frame)
		end
	end
	
	scanObjects("CharacterBag%dSlot")
	scanObjects("BT4Button%d")
	scanObjects("BT4StanceButton%d")
	scanObjects("LunarMenu%dButtonHeader")
	scanObjects("MacaroonButton%d")
	
	if( self.db.collisionObjects.Minimap ) then frameList[Minimap] = true end
	if( self.db.collisionObjects.MicroButtons ) then
		frameList[CharacterMicroButton] = true
		frameList[SpellbookMicroButton] = true
		frameList[TalentMicroButton] = true
		frameList[AchievementMicroButton] = true
		frameList[QuestLogMicroButton] = true
		frameList[SocialsMicroButton] = true
		frameList[PVPMicroButton] = true
		frameList[LFDMicroButton] = true
		frameList[MainMenuMicroButton] = true
		frameList[HelpMicroButton] = true
	end

	if( self.db.collisionObjects.UnitFrames ) then
		for frame in pairs(ClickCastFrames) do
			frameList[frame] = true
		end
	end
	
	local miscMods = {Recount_DetailWindow, OmenAnchor, PROptions_Config, AtlasLootDefaultFrame}
	for _, frame in pairs(miscMods) do
		frameList[frame] = true
	end
	
	if( AutoBar and AutoBar.buttonList ) then
		for _, button in pairs(AutoBar) do
			frameList[button] = true
		end
	end
		
	if( NinjaPanel and NinjaPanel.panels ) then
		for i=1, #(NinjaPanel.panels) do
			frameList[NinjaPanel.panels[i]] = true
		end
	end
end

-- Not the cleanest solution to finding a valid frame, but it reduces the chance of it having to look through a new valid frame if we cache a list of valid ones
local screenThreshold = GetScreenHeight() - 175
function LetItSnow:BuildValidFrames()
	self:CreateList()
	table.wipe(validFrames)
	
	for frame in pairs(frameList) do
		if( frame:IsVisible() ) then
			if( verifiedPosition[frame] ) then
				for i=1, verifiedPosition[frame] do
					table.insert(validFrames, frame)
				end
			-- Make sure that there is at least 175 pixels between the top of the UIObject and the screen to prevent odd looking drifts
			elseif( not frame.GetTop or not frame:GetTop() or frame:GetTop() <= screenThreshold ) then
				-- Try and give wider frames a larger chance of being chosen to balance out the fact that their are less of them
				verifiedPosition[frame] = math.ceil(frame:GetWidth() / 20) + math.ceil(frame:GetTop() / 100)
				for i=1, verifiedPosition[frame] do
					table.insert(validFrames, frame)
				end
			end
		end
	end
end

function LetItSnow:FindLanding()
	local valid = #(validFrames)
	if( valid <= 1 ) then return nil end
	
	for i=1, 5 do
		local frame = validFrames[math.random(1, valid)]
		if( frame:IsVisible() ) then
			return frame
		end
	end
end

-- Start a new flake off
function LetItSnow:GenerateFlake()
	local flakeSize = self.db.flakeSize + math.random(-self.db.sizeRandomizer, self.db.sizeRandomizer)

	local frame = getFlake()
	frame:SetHeight(flakeSize)
	frame:SetWidth(flakeSize)
	frame:SetAlpha(self.db.flakeAlpha)
	frame.flake:SetTexture(SNOWFLAKES[math.random(1, #(SNOWFLAKES))])
	frame:Show()

	-- Figure out what side the flake is going to start on
	local halfWidth = GetScreenWidth() / 2
	local xOffset = math.random(-halfWidth, halfWidth)
	frame:ClearAllPoints()
	frame:SetPoint("TOP", UIParent, "TOP", xOffset, flakeSize)
	
	-- Figure out the coords of the spot we are landing on
	local landFrame = self.db.collision and self:FindLanding()
	local endX, endY
	if( landFrame ) then
		local landX, landY = landFrame:GetCenter()
		local frameX, frameY = frame:GetCenter()
		local frameScale, landScale = frame:GetEffectiveScale(), landFrame:GetEffectiveScale()
		endX = (landX * landScale) - (frameX * frameScale)
		endY = (landY * landScale) - (frameY * frameScale) + (landFrame:GetHeight() / 2)
		
		-- Figure out the allowed variance in landing area, it should not exceed the frame we are landing on, and it should not exceed the flakes width
		local landingDrift = (landFrame:GetWidth() / 2) - flakeSize
		endX = endX - math.random(-landingDrift, math.abs(landingDrift))
	else
		startX = 512
		screenWidth = 1024
		
		endX = math.random(-xOffset, math.max(0, GetScreenWidth() - math.abs(xOffset))) 
		endY = -GetScreenHeight()
	end
		
	frame.fallAnim:SetDuration(self.db.fallDuration + math.random(-self.db.fallDurRandomizer, self.db.fallDurRandomizer))
	frame.fallAnim:SetOffset(endX, endY)
	frame.fallAnim:SetSmoothing("IN")
	frame.fallAnim:SetOrder(1)
		
	-- Fade out after it finishes moving
	frame.fadeAnim:SetDuration(self.db.fadeDuration + math.random(-self.db.fadeDurRandomizer, self.db.fadeDurRandomizer))
	frame.fadeAnim:SetChange(-1)
	frame.fadeAnim:SetOrder(2)
	frame.fadeAnim:SetStartDelay(1)
	frame.fadeAnim:SetSmoothing("IN")
			
	-- Sets the flakes to rotation 360 degrees clockwise if they are starting from the left side, counterclockwise if they start on the right
	frame.rotateAnim:SetDuration(math.random(10, 30))
	frame.rotateAnim:SetDegrees(math.random(1, 2) == 2 and 360 or -360)
	frame.rotateAnim:SetOrder(1)
	
	frame.rotateGroup:Play()
	frame.animGroup:Play()
end

-- Stopping flakes, either instantly or a gradual fade out
function LetItSnow:QuickStop()
	self:CancelAllTimers()

	for _, frame in pairs(activeFlakes) do
		frame.suppress = true
		frame.animGroup:Stop()
		frame:Hide()
	end
	
	for i=#(activeFlakes), 1, -1 do
		local frame = table.remove(activeFlakes, i)
		frame.suppress = nil
		table.insert(inactiveFlakes, frame)
	end
end

function LetItSnow:GradualStop()
	self:CancelAllTimers()

	for _, frame in pairs(activeFlakes) do
		if( not frame.quickFadeGroup ) then
			frame.quickFadeGroup = frame:CreateAnimationGroup()
			frame.quickFadeAnim = frame.quickFadeGroup:CreateAnimation("Alpha")
			frame.quickFadeAnim:SetScript("OnFinished", OnFinished)
			frame.quickFadeAnim.parent = frame
		end
		
		frame.quickFadeAnim:SetDuration(0.75)
		frame.quickFadeAnim:SetChange(-1)
		frame.quickFadeAnim:SetSmoothing("IN")
		frame.quickFadeAnim:SetOrder(1)
		frame.quickFadeGroup:Play()
	end
end

-- Deals with creating snow, this creates it in 30 second intervals
function LetItSnow:ScheduleFall()
	scheduledFall = nil
	self:CancelTimer("ScheduleLightSnow", true)
	self:CancelTimer("ScheduleHeavySnow", true)
	
	if( self.db.snowType == "light" ) then
		self:ScheduleTimer("ScheduleLightSnow", self.db.minFallInterval + math.random(0, self.db.fallRandomizer))
	else
		self:ScheduleTimer("ScheduleHeavySnow", self.db.minFallInterval + math.random(0, self.db.fallRandomizer))
	end
end

function LetItSnow:ScheduleLightSnow()
	if( not hasSnowflakesBuff ) then
		scheduledFall = true
		LetItSnow:BuildValidFrames()

		for i=1, 50 do
			local buff = math.ceil(i * 0.05)
			self:ScheduleTimer("GenerateFlake", buff + math.random(0, self.db.fallDuration))
		end
	end
	
	self:ScheduleFall()
end

function LetItSnow:ScheduleHeavySnow()
	if( not hasSnowflakesBuff ) then
		scheduledFall = true
		LetItSnow:BuildValidFrames()
		
		for i=1, 100 do
			local buff = math.ceil(i * 0.20)
			self:ScheduleTimer("GenerateFlake", buff + math.random(0, self.db.fallDuration))
		end
	end
		
	self:ScheduleFall()
end

local SNOWFLAKES_BUFF = GetSpellInfo(44755)
function LetItSnow:UNIT_AURA(event, unit)
	if( unit ~= "player" ) then return end

	local hasSnow = UnitBuff("player", SNOWFLAKES_BUFF)
	if( hasSnow and not hasSnowflakesBuff and not scheduledFall ) then
		LetItSnow:BuildValidFrames()
		for i=1, 200 do
			self:ScheduleTimer("GenerateFlake", i)
		end
	end

	hasSnowflakesBuff = hasSnow
end

-- Mod enabling/disabling
function LetItSnow:Enable()
	if( self.enabled ) then return end
	self.enabled = true
	
	self:ScheduleFall()
	self:UNIT_AURA()
	self:RegisterEvent("UNIT_AURA")
end

function LetItSnow:Disable()
	if( not self.enabled ) then return end
	self.enabled = nil
	
	self:UnregisterEvent("UNIT_AURA")
	self:GradualStop()
end

function LetItSnow:PLAYER_REGEN_DISABLED()
	isInCombat = true
	self:CheckStatus()
end

function LetItSnow:PLAYER_REGEN_ENABLED()
	isInCombat = nil
	self:CheckStatus()
end

function LetItSnow:CheckStatus()
	if( ( not self.db.inCombat and isInCombat ) or ( not self.db.whileGrouped and ( GetNumRaidMembers() > 0 or GetNumPartyMembers() > 0 ) ) ) then
		self:Disable()
	else
		self:Enable()
	end
end

function LetItSnow:WatchEvents()
	if( not self.db.inCombat ) then
		self:RegisterEvent("PLAYER_REGEN_DISABLED")
		self:RegisterEvent("PLAYER_REGEN_ENABLED")
	else
		self:UnregisterEvent("PLAYER_REGEN_DISABLED")
		self:UnregisterEvent("PLAYER_REGEN_ENABLED")
	end
			
	if( not self.db.whileGrouped ) then
		self:RegisterEvent("PARTY_MEMBERS_CHANGED")
		self:RegisterEvent("RAID_ROSTER_UPDATE")
	else
		self:UnregisterEvent("PARTY_MEMBERS_CHANGED")
		self:UnregisterEvent("RAID_ROSTER_UPDATE")
	end
end

function LetItSnow:ADDON_LOADED(event, addon)
	if( addon ~= "LetItSnow" ) then return end

	LetItSnowDB = LetItSnowDB or {
		inCombat = false,
		onHandfulSnowflakes = true,
		whileGrouped = true,
		collision = true,
		snowType = "light",
		minFallInterval = 60,
		fallRandomizer = 60,
		flakeAlpha = 0.70,
		flakeSize = 20,
		fadeDuration = 3,
		fadeDurRandomizer = 2,
		sizeRandomizer = 4,
		fallDuration = 25,
		fallDurRandomizer = 5,
	}
	
	if( LetItSnowDB.collision == nil ) then LetItSnowDB.collision = true end
	LetItSnowDB.fadeDuration = LetItSnowDB.fadeDuration or 3
	LetItSnowDB.fadeDurRandomizer = LetItSnowDB.fadeDurRandomizer or 2
	LetItSnowDB.collisionObjects = LetItSnowDB.collisionObjects or {["MultiBarLeftButton%d"] = true, ["MultiBarRightButton%d"] = true, ["MultiBarBottomRightButton%d"] = true, ["MultiBarBottomLeftButton%d"] = true, ["MultiCastActionButton%d"] = true, ["ShapeshiftButton%d"] = true, ["PetActionButton%d"] = true, ["BonusActionButton%d"] = true, ["ActionButton%d"] = true, ["VehicleMenuBarActionButton%d"] = true, ["UnitFrames"] = true, ["MicroButtons"] = true, ["Minimap"] = true, ["ChatFrame%d"] = true}
	
	self.db = LetItSnowDB
	
	self.PARTY_MEMBERS_CHANGED = self.CheckStatus
	self.RAID_ROSTER_UPDATE = self.CheckStatus

	self:WatchEvents()
	self:CheckStatus()
end


LetItSnow:RegisterEvent("ADDON_LOADED")
LetItSnow:SetScript("OnEvent", function(self, event, ...)
	LetItSnow[event](LetItSnow, event, ...)
end)
